function [verifiedSrcPoints, verifiedDstPoints, t, dstPatchInfo, matches, scores] = REX(srcFile, dstFile, varargin )
% MAIN function for Region EXpansion (REX)
% 
% input:   srcFile, dstFile - strings for source and destination image locations
%          optional inputs: 
%               1. 'secondBestFactor' - see details in "getAffineAdaptedMatches" or in "getMSERMatches"                
%               2. 'doThrowRedundantMatches' - A boolean flag that controls whether we throw phase-zero matches that are too similar to each other (saves signficant computaional time)
%               3. 'phaseZeroMethod': A string indicating the method used to detect phase-zero matches. compatible options: {'HARRISAFFINE','MSER'}
%               4. 'numberOfFurtherExpansions'- A scalar for how many times to try and expand a region further after it already suceeded to expand (for planar scenes a high number may prove effective).
%               5. 'generateOutput' - a boolean flag that determines whether to save output files (both text files and images)
%
%  
% output:  verifiedSrcPoints, verifiedDstPoints - two [NX2] matrices
%                                                 containing oredered matches between source and destination
%                                                 respectively
%          t - processing time for each step
% 
% if generateOutput = true
%          'intermediateMatches.js' - phase zero matches, before expansion
%                              arranged as <Xsrc_i> <Ysrc_i> <Xdst_i> <Ydst_i> 
%                              where (Xsrc_i, Y_src_i) is a point in the
%                              source image which matches (Xdst_i, Y_src_i)
%                              in the destination image
% 
%          'verifiedMatches.js' - final matches, after expansion, arranged
%                                  in the same way as 'intermediateMatches.js'
% 
% modes of operation for REX
% --------------------------
% 
% REX(srcFile, dstFile) - find correspondences between srcFile and dstFile
%  
% REX(srcFile, dstFile, 'secondBestFactor', value) - sift's second best threshold. Rejecting matches if second 
%                                                    best is too close to the first one as suggested by D.Lowe (default - 1.5)
% 
% REX(srFile, dstFile, 'doThrowRedundantMatches', value) - Throwing
% phase-zero matches that are too similar to another match (thus redundant)
% 
% 
warning('off', 'MATLAB:nearlySingularMatrix');                  %%%% maybe treat the warning with respect, instead of shutting it down? (same "scale" tau and A would meet the criterion)
addpath(genpath('controlSection'));
addpath(genpath('homography'));
addpath(genpath('normxcorr2MEX'));
addpath(genpath('other'));
addpath(genpath('phaseOne'));
addpath(genpath('phaseZero'));
addpath(genpath('rotationEstimation'));
addpath(genpath('vlfeatMEX'));

%% parser
algoParse = tic;
params = parseInput(varargin{:});
% disp(params);
fprintf('Parsing input arguments.\n');

if (nargin > 2)
    if isfield(params, 'secondBestFactor')
        m1 = getfield(params, 'secondBestFactor');              %% user input
    else
        m1 = 1.5;                                     %% default value
    end

    if isfield(params, 'generateOutput')
        m2 = getfield(params, 'generateOutput');
    else
        m2 = false;
    end
    
    if isfield(params, 'doThrowRedundantMatches')
        m3 = getfield(params, 'doThrowRedundantMatches');
    else
        m3 = true;
    end
    if isfield(params, 'phaseZeroMethod')
        m4 = getfield(params, 'phaseZeroMethod');
    else
        m4 = 'MSER';
    end
    if isfield(params, 'numberOfFurtherExpansions')
        m5 = str2num(getfield(params, 'numberOfFurtherExpansions'));
    else
        m5 = 3;
    end
else
    m1 = 1.5;
    m2 = false;
    m3 = true;
    m4 = 'MSER';
    m5 = 3;
end

secondBestFactor = m1;
isDebug = m2;
doThrowRedundantMatches = m3;
switch (m4)
    case 'MSER'
        phaseZeroMatcher = @getMSERMatches;
    case 'HARRISAFFINE'
        phaseZeroMatcher = @getAffineAdaptedMatches;
end
numberOfFurtherExpansions = m5;
tParser = toc(algoParse);
fprintf('Algorithm initialization.\n');
algoInit = tic;

srcImg = imread(srcFile);
dstImg = imread(dstFile);

if (size(srcImg,3) == 3)
    srcImg = rgb2gray(srcImg);
end
if (size(dstImg,3) == 3)
    dstImg = rgb2gray(dstImg);
end
%EnableMultithreading;
%if(isempty(gcp('nocreate')))
%   parpool 
%end
tInit = toc(algoInit);
%% phase ZERO
phaseZero = tic;
fprintf('Processing Phase 0: Match generation.\n');
% using phase-zero matcher , source and destination normalizing matrices are
% acquired (srcNStack, dstNStack repectively) 
[srcPatchInfo, dstPatchInfo, matches, scores, srcPhaseZero, dstPhaseZero] = phaseZeroMatcher(srcImg, dstImg, secondBestFactor);
if (doThrowRedundantMatches)
    [~,qualitySortedIdxs] = sort(scores,'Ascend');
    srcNStack = reshape([srcPatchInfo(matches(1,qualitySortedIdxs)).N],3,3,length(matches));
    dstNStack = reshape([dstPatchInfo(matches(2,qualitySortedIdxs)).N],3,3,length(matches));
    srcPhaseZero = srcPhaseZero(qualitySortedIdxs,:);dstPhaseZero = dstPhaseZero(qualitySortedIdxs,:);
    if exist('uniquetol','builtin') %matlab 2015a and up
        [ selectedIdxs ] = throwRedundantMatches( srcPhaseZero, dstPhaseZero, srcNStack, dstNStack, 20);
    else
        warning('uniquetol does not exist in your matlab version, expanding using all phase-zero matches (will work slower than planned!)');
        selectedIdxs = 1:length(srcPhaseZero);
    end
    srcPhaseZero = srcPhaseZero(selectedIdxs,:);dstPhaseZero = dstPhaseZero(selectedIdxs,:);
    srcNStack =  srcNStack(:,:,selectedIdxs);dstNStack =  dstNStack(:,:,selectedIdxs);
else
    srcNStack = reshape([srcPatchInfo(matches(1,:)).N], 3, 3, length(matches));
    dstNStack = reshape([dstPatchInfo(matches(2,:)).N], 3, 3, length(matches));
end
% output phase-zero matches as ascii txt file
PhaseZeroMatches = double([srcPhaseZero, dstPhaseZero]);
if (isDebug)
    saveMatchesToFile('intermediateMatches.js', PhaseZeroMatches, true, size(srcImg,2));
end



tZERO = toc(phaseZero);

%% phase ONE
phaseOne = tic;
fprintf('Processing Phase 1: Region Expansion.\n');
% once the initial, phase ZERO, points have been matched, they are expanded
% in a local affine region in order to match more points - A.K.A. phase ONE
% points
[expandersList] = expandAllRegions(srcNStack, dstNStack,  srcImg, dstImg, true, numberOfFurtherExpansions);                             %%%% should replace 0->isDebug
[ verifiedSrcPoints, verifiedDstPoints, ~, ~] = expandersListToPointMatches( expandersList );
verifiedMatches = [verifiedSrcPoints, verifiedDstPoints];
if (isDebug)
    saveMatchesToFile('verifiedMatches.js', verifiedMatches, false, size(srcImg,2));
end

tONE = toc(phaseOne);

%% Draw and Print Matches
tic;
if (isDebug)
    fprintf('Output generation. ');
    drawAndPrintMatches(imread(srcFile), imread(dstFile), [srcPatchInfo.location], [dstPatchInfo.location], [matches(1,:); matches(2,:)],'intLinesFig.png', false);
    drawAndPrintMatches(imread(srcFile), imread(dstFile), verifiedSrcPoints', verifiedDstPoints', [1:size(verifiedSrcPoints, 1); 1:size(verifiedDstPoints, 1)],'rexLinesFig.png', true);
end
tOutput = toc;

%% Finish
fprintf('Done.\n');
t = [tParser, tInit, tZERO, tONE, tOutput];
fprintf('Algorithm finished in %.2fs.\nTimes: \nInput Parsing: %.2fs\nAlgo Init: %.2fs\nPhase Zero: %.2fs\nPhase One: %.2fs\nOutput generation: %.2fs\n',sum(t), tParser, tInit, tZERO, tONE, tOutput);



%% parser function
function [params] = parseInput(varargin)

if (mod(size(varargin),2) ~= 0)
    error('wrong input');
end

params = struct();
for ii= 1:2:size(varargin, 2)
    tempProp = varargin{ii};
    tempVal = varargin{ii+1};
    params = setfield(params, tempProp, tempVal);
end
end

end
