classdef regionExpanderDeltaBased < handle
% 
% This class contatains "the inside" of Region EXpansion (REX) algorithm
% 
% It conatains most of the methods for expansion, divided into two main
% sections - initial and further expansions (see footnote)
% 
% There are several different methods to construct an object, including a
% preallocating one (works also on matrix of objects)
% 
% 
% footnote
% --------
% As we improve the estimation for the spatial transformation, we can use a
% more accurate analysis (marked as 'further') in order to achieve more 
% accurate expansion points, as well as less computation time. 
% 
    
    
    properties
        
        %% Expander(DEBUG) props.
        Hhat                                % improved estimated affine matrix between the affine region at srcImg and dstImg
        Hprevious                           % temporary result for Hhat's last iteration estimation
        Hinitial                            % estimated affine matrix given by phase ZERO
        invHerrCovariance                   % The inverse of the (estimated) error covariance matrix in estimating Hhat (6X6)                                                                                       % 
        currentCorrespondences              % A struct containing phaseOne matches and their properties 
                                                % .src - the phase-one points in the source image
                                                % .dst- the corresponding phase-one points in the destination image
                                                % qualities- the qualities(correlations) of each phase-one correspondence
                                                % .covMats - The phase-one confidence matrices (an estimated error covarince matrix for each correspondences)
                                                % .invCovMats - The respective inverses of covMats
        srcPrincipalAxes                    % source region principal axes (axes of the fitted ellipse)
        dstPrincipalAxes                    % destination region principal axes
        area                                % region's area in pixels, computed by principal axis as an ellipse 
        srcPtsCentroid                      % source region centroid (destination region centroid can be easily computed using Hhat)
        wasRejected = false;                % logic variable for marking a rejected region
        wasStopped = false;                 % logic variable for marking a region that finished its expansion (and will not continue expanding)
        wasExpanded = false;                % logic variable for marking a region that was exapanded and approved
        initialExpansionFactor = 1.5;       % scaling factor for initial expansion
        downSamplingFactor;                 % A variable used to control the spatial density of new samples (samples are roughly within "downSamplingFactor" from each other)                                                                                       %
        minimalInitialSamples;              % lower bound for minimal angular intervals for initial circular sampling 
        phaseZeroId;                        % expanders' ID number. debug purposes
        expansionCtr = 0;                   % counts how many times we tried to expand this expander (for debug)
    end
    
    properties(Constant)

        minimalAreaGrowthFactor = 0.5;                  %The minimal allowed rate of a desired expansion factor - when breached, expansion will stop
        maximalRelativePredictionErr = 0.2;             %The maximal error per-pixel of prediction distance
        downSamplingDivider = 2;                        %The distance between new phase-one samples in "templateExtent" units
        minimalMinorAxisLength = 20;                    %The minimal acceptable minor axis length of a region            
        minimalSearchWindowSlack = 3;                   %minimal side length for the correlation window
        maximalSearchWindowSlack = 16;                  %maximal side length for the correlation window
        templateExtent = 8;                             %the distance between the template center and any of the sides (square shaped templates)
        correlationDeviationTsh = 0.7                   %How much deviation allowed from the maximal detected correlation
        minimalCorreltionTsh = 0.5;                     %the minimal acceptable value for the maximal NCC result
        timesSigmaPredictionSlackFactor = 2;            %The mahalnobis distance allowed for determining the allowed prediction error
        minimalThetaSampleRes = 3;                      %minimum number of samples in the initial sample circle
        maximalNumberOfInitialSamples = 30;             %maximum number of samples per initial expansion
        maximalTransformInflation = 50;                 %a boundary on the scaling of a given affine transformation           
    end
    
    methods
        %% Constructor
        function [expander] = regionExpanderDeltaBased(varargin)            
            if nargin == 1 %preallocating constructor
                expander.currentCorrespondences.src = [];
                expander.currentCorrespondences.dst = [];
                expander.currentCorrespondences.qualities = [];
                expander.currentCorrespondences.covMats = [];
                expander.currentCorrespondences.invCovMats = [];
                expander = repmat(expander,[varargin{:}]);
                
            else
                if nargin == 4
                    expander.srcPtsCentroid = varargin{1};
                    expander.srcPrincipalAxes = varargin{2};
                    expander.Hhat = varargin{3};
                    expander.dstPrincipalAxes = expander.Hhat(1:2,1:2)*expander.srcPrincipalAxes;
                    expander.phaseZeroId = varargin{4};
                else
                    H0 = varargin{1};
                    HT = varargin{2};
                    expander.Hhat = HT\H0;
                    [expander.srcPtsCentroid,v1,v2] = getRegionPrincipleAxes([(doInverseAffineTransform([0 -1 ; 0 1], H0));(doInverseAffineTransform([-1 0 ; 1 0], H0))]);
                    expander.srcPrincipalAxes(:,1) = v1/sqrt(2);%divide by sqrt(2) to get the ellipse axis length
                    expander.srcPrincipalAxes(:,2) = v2/sqrt(2);%divide by sqrt(2) to get the ellipse axis length
                    [~,v1,v2] = getRegionPrincipleAxes([(doInverseAffineTransform([0 -1 ; 0 1], HT));(doInverseAffineTransform([-1 0 ; 1 0], HT))]);
                    expander.dstPrincipalAxes(:,1) = v1/sqrt(2);%divide by sqrt(2) to get the ellipse axis length
                    expander.dstPrincipalAxes(:,2) = v2/sqrt(2);%divide by sqrt(2) to get the ellipse axis length
                    expander.area = pi*norm(expander.srcPrincipalAxes(:,1))*norm(expander.srcPrincipalAxes(:,2));%ellispe area
                    if (nargin == 4)
                        expander.phaseZeroId = varargin{3};
                    end
                end
                if (det(expander.Hhat(1:2,1:2)) > expander.maximalTransformInflation || det(expander.Hhat(1:2,1:2)) < 1/expander.maximalTransformInflation || norm(expander.srcPrincipalAxes(:,2)) < 5)
                    expander.wasStopped = true;
                else
                    % produce a rough evaluation of the inverse error
                    % covariance matrix of the affine transformation. The
                    % fixed numbers here were not carefully studied from
                    % statistics, but are reasonable with respect to our
                    % expertiments:
                    correspondenceErrInvCov = blkdiag(eye(2), 0.25*eye(2), 0.25*eye(2), 0.25*eye(2), 0.25*eye(2));%0.25 is for 2 pixel error std, 0.11 is for 3 pixel error
                    [ expander.invHerrCovariance ] = 0.11*getInvEstimationCovariance(repmat(expander.srcPtsCentroid, 5, 1) + [0 0;expander.srcPrincipalAxes';-expander.srcPrincipalAxes'], correspondenceErrInvCov);%0.11 is for multiplying the std by 3
                    expander.Hinitial = expander.Hhat;
                    expander.Hprevious = expander.Hhat;
                    expander.currentCorrespondences.src =...
                        repmat(expander.srcPtsCentroid, 3, 1) + [[0 0];expander.srcPrincipalAxes'];
                    expander.currentCorrespondences.dst = doAffineTransform(expander.currentCorrespondences.src, expander.Hhat);
                    expander.currentCorrespondences.qualities = 0.5*ones(3, 1);%for this initial stage, we don't really now the qualities
                    expander.currentCorrespondences.covMats = repmat(100*eye(2), 1, 1, 3);%for this initial stage, we don't really now the error covariances for each phase-one match
                    expander.currentCorrespondences.invCovMats = repmat((1/100)*eye(2), 1, 1, 3);%for this initial stage, we don't really now the error covariances for each phase-one match
                    expander.downSamplingFactor = 1/((expander.templateExtent)*regionExpanderDeltaBased.downSamplingDivider);
                    expander.minimalInitialSamples = max(ceil(pi*(regionExpanderDeltaBased.minimalMinorAxisLength*expander.downSamplingFactor)^2), 4);%we want a prescribed density, but at least 4 samples
                end
            end
        end %%Constructor
        
    end
    
    
    %peripherial private methods
    methods (Access = private)
        %% scanning / correlation related methods
        

        function [outPixels, matchQualities, oobIndicators, errorCorrMats, isGarbageIndicators, predictabilityCovs, initialCorrelations] = scanForMatches(this, dstImg, templates, dstPts, srcPts, isDebug)
            % scanForMatches - search for new phase-one matches
            % Inputs:
            %   dstImg - destination image
            %   templates - templates for the NCC scans
            %   dstPts - predictions of the corresponding points to srcPts in
            %   the destination image (centers of scanning windows)
            %   srcPts - the samples in the source image
            %   isDebug - binary flag
            % Outputs:
            %   outPixels - the corrected locatins in the destination image
            %   matchQualities - correspondes qualities (correlations)
            %   oobIndicators - binary indicators for matches being out of image bounds
            %   errorCorrMats - estimated error correlation matrix of each located match
            %   isGarbageIndicators- Indicator to if match should be discarded
            %   predictabilityCovs - estimated error correlation matrix in
            %   predicting the location of each point
            %   initialCorrelations - correlation of each predicted location
            matchQualities = zeros(length(templates), 1);
            initialCorrelations = zeros(length(templates), 1);
            errorCorrMats = zeros(2, 2, length(templates));
            outPixels = zeros(length(templates), 2, 'single');
            oobIndicators = false(1, length(templates));
            isGarbageIndicators = false(1, length(templates));
            XT = zeros(2*size(srcPts, 1), 6);
            XT(1:2:end,1:3) = [srcPts ones(size(srcPts, 1), 1)];
            XT(2:2:end,4:6) = [srcPts ones(size(srcPts, 1), 1)];
            predictabilityCovs = zeros(2, 2, length(templates));
            for k = 1:length(templates)
                dstPt = dstPts(k,:);
                predictabilityCovs(:,:,k) = ((XT((2*k-1):2*k,:))/this.invHerrCovariance*(XT((2*k-1):2*k,:)')) ;          %%%% inv
                [oobIndicators(k), outPixels(k,:), matchQualities(k), errorCorrMats(:,:,k),isGarbageIndicators(k),initialCorrelations(k)] = this.findMatch(dstImg, dstPt, predictabilityCovs(:,:,k), templates{k}, isDebug);
            end
        end
        
        function [isOOB, outPixel, matchQuality, errorCorrMat, isGarbage,initialCorrelation] = findMatch(this, dstImg, centerPnt, predictabilityCovariance, dstRoi, isDebug)
            % findMatch - search for a new phase-one match
            % Inputs / Outputs: see method "scanForMatches"
            isOOB = false;
            xSlack = sqrt(predictabilityCovariance(1,1)) * regionExpanderDeltaBased.timesSigmaPredictionSlackFactor;
            ySlack = sqrt(predictabilityCovariance(2,2)) * regionExpanderDeltaBased.timesSigmaPredictionSlackFactor;
            searchWindowSlack = max([ [xSlack ySlack];[this.minimalSearchWindowSlack this.minimalSearchWindowSlack]]);
            [subImage, offset] = regionExpanderDeltaBased.cropSubImage(dstImg, centerPnt, [this.templateExtent this.templateExtent] + searchWindowSlack, this.templateExtent + this.minimalSearchWindowSlack);
            if (size(subImage, 1) < size(dstRoi, 1) || size(subImage, 2) < size(dstRoi, 2))
                outPixel = centerPnt;
                isOOB = true;
                isGarbage = true;
                matchQuality = 0;
                errorCorrMat = inf(2);
                initialCorrelation = 1;
                return
            end
            [cords, matchQuality, errorCorrMat, isGarbage, ~, initialCorrelation] = patchedFilterDelta(single(subImage), dstRoi, isDebug, regionExpanderDeltaBased.minimalCorreltionTsh);
            outPixel = cords + offset;
        end
        
        function [correlations,badTemplateInd] = getEstimationInducedCorrelation(this, srcImg, dstImg)
            % getEstimationInducedCorrelation - get correlations between current source points and target points 
            % Inputs:
            %   srcImg - source image
            %   dstImg - destination image
            % Outputs: 
            %   correlations - the correlations considering the estimated affine warp
            %   badTemplateInd - indicator for bad locations
            [srcTemplates,~,badTemplateInd] = getTemplates(srcImg, this.currentCorrespondences.src, this.templateExtent, this.Hhat);
            srcTemplates = srcTemplates(~badTemplateInd);
            dstTemplates = getEyeTemplates(dstImg, doAffineTransform(this.currentCorrespondences.src(~badTemplateInd,:), this.Hhat), this.templateExtent);
            correlations = arrayfun(@(k) corr2(srcTemplates{k}(:),dstTemplates{k}(:)), 1:sum(~badTemplateInd));
        end
        
    end
    %% public methods
    methods (Access = public)
        
        function [srcPoints, dstPoints, qualities] = getTopMatches(this, percentage)
            % getTopMatches - get a precentage of current matches of this
            % expander and their qualities
            [q,idxs] = sort(this.currentCorrespondences.qualities);
            selectionLength = round(percentage*length(idxs));
            selection = idxs(1:selectionLength);
            srcPoints = this.currentCorrespondences.src(selection,:);
            dstPoints = this.currentCorrespondences.dst(selection,:);
            qualities = q(selection);
        end
        
        function [] = visualizeCurrentState(this,srcImg,dstImg)
             % visualizeCurrentState - visualization of the state of the expander
             % Inputs:
             %   srcImg - source image
             %   dstImg - destination image
             % Outputs:
             %   correlations - the correlations considering the estimated affine warp
             %   badTemplateInd - indicator for bad locations
            if (~isempty(this.currentCorrespondences.src))
                figure(100);imshow(srcImg)
                imshowAndMarkLandMarks(srcImg, this.currentCorrespondences.src, [this.templateExtent this.templateExtent], 100);
                regionExpanderDeltaBased.textifyMyPoints(this.currentCorrespondences.src, [1 1 1]);
                figure(101);imshow(dstImg)
                imshowAndMarkLandMarks( dstImg, this.currentCorrespondences.dst, [this.templateExtent this.templateExtent], 101);
                regionExpanderDeltaBased.textifyMyPoints(this.currentCorrespondences.dst, [1 1 1]);
            end
        end
        
        %% 
        function [] = expand(this,srcImg,dstImg,srcCornerImg, expansionFactor, isDebug)
            % expand - expand the region of this expander
            % Inputs:
             %  srcImg - source image
             %  dstImg - destination image
             %  srcCornerImg - corrnerness map of the source image
             %  expansionFactor - how much to expand in every axis
             %  isDebug - a boolean flag  
            this.expansionCtr = this.expansionCtr + 1;
            if (this.wasExpanded)
                this.furtherExpand(srcImg, dstImg, srcCornerImg, expansionFactor, isDebug)
            else
                this.initialExpansion(srcImg,dstImg, srcCornerImg, this.initialExpansionFactor, isDebug)
            end
            
        end
        
    end
    
    %% expansion control related private methods
    methods (Access = private)
        %% initial expansion
        function [] = initialExpansion(this, srcImg, dstImg, srcCornerImg, expansionFactor, isDebug)    
            % initialExpansion - expansion of yet to be verified region
            % Inputs:
             %  srcImg - source image
             %  dstImg - destination image
             %  srcCornerImg - corrnerness map of the source image
             %  expansionFactor - how much to expand in every axis
             %  isDebug - a boolean flag  
            [isSuccess, dstRois, dstPts, srcPts, ~] = this.prepareInitialSamples(srcImg,srcCornerImg,expansionFactor);      %%%%actualExpansionFactor is not used
            if (~isSuccess)
                this.wasStopped = true;
                return;
            end
            [dstPts, matchQualities, ~, errorCorrMats, isGarbageIndicators, ~, initialCorrelations] = this.scanForMatches(dstImg, dstRois, dstPts, srcPts, isDebug);
            srcPts(isGarbageIndicators,:) = [];
            dstPts(isGarbageIndicators,:) = [];
            matchQualities(isGarbageIndicators) = [];
            errorCorrMats(:,:,isGarbageIndicators) = [];
            initialCorrelations(isGarbageIndicators) = [];
            goodQualitiesInd = matchQualities >  max(matchQualities)*this.correlationDeviationTsh;
            
            if ( sum(goodQualitiesInd) >= 4)
                srcPts = srcPts(goodQualitiesInd,:);
                dstPts = dstPts(goodQualitiesInd,:);
                errorCorrMats = errorCorrMats(:,:,goodQualitiesInd);
                matchQualities = matchQualities(goodQualitiesInd);
                initialCorrelations = initialCorrelations(goodQualitiesInd);
                [goodIdxs, ~] = this.doGeometricAnalysisInitial(srcPts, dstPts, matchQualities, errorCorrMats);
                
                if(this.wasExpanded)
                    initialCorrelations = initialCorrelations(goodIdxs);
                    [currentcorrelations,badTemplateInd] = this.getEstimationInducedCorrelation(srcImg, dstImg);
                    initialCorrelations = initialCorrelations(~badTemplateInd);
                    improvementIndicator = (currentcorrelations > 1.05*initialCorrelations' | currentcorrelations > 0.9) & currentcorrelations > this.minimalCorreltionTsh;
                    if ( (mean(improvementIndicator) < 0.8) && (sum(currentcorrelations)/sum(initialCorrelations) < 1.05)  || sum(improvementIndicator) < 4)
                        this.wasExpanded = false;
                        this.wasRejected = true;
                        this.wasStopped = true;
                    end
                end
                
                
            else
                this.wasRejected = true;
                this.wasStopped = true;
            end
        end
        function [chosenIdxs, Hhat] = doGeometricAnalysisInitial(this, srcPts, dstPts, qualities, errCovs)       
            % doGeometricAnalysisInitial - check whether the given matches agree on an affine model and are enough for verfication 
            % Inputs:
             %  srcPts - the points in the source image
             %  dstPts - corresponding points in the destination image
             %  qualities - match qualites (correlations)
             %  errCovs - error covariance matrices of the matches
            errorDets = zeros(size(srcPts, 1), 1);
            invErrCovs = zeros(size(errCovs));
            qualityFactors = (qualities / max(qualities)).^-2;
            for k = 1:size(srcPts, 1)
                errCovs(:, :, k) = qualityFactors(k) * errCovs(:, :, k);
                [invErrCovs(:, :, k), errorDets(k)] = inv2By2Mat(errCovs(:, :, k));
                %             invCovMat((2*k-1):(2*k), (2*k-1):(2*k)) = invErrCovs(:, :, k);
            end
            [Hhat, ~, predictabilityCovs, ~] = weighteAffineEstimation(srcPts, dstPts, invErrCovs, srcPts);
            [ ~, mahalanobisErrs ] = evalWeightedAffineEstimation(Hhat, predictabilityCovs, srcPts, dstPts, errCovs);
            goodIdxs = mahalanobisErrs < (this.timesSigmaPredictionSlackFactor)/2;
            if (all(goodIdxs))
                chosenIdxs = 1:size(srcPts, 1);
                isSucsses = true;
            else if sum(goodIdxs) >= 4
                    chosenIdxs = find(goodIdxs);
                    [Hhat, ~, predictabilityCovs, ~] = weighteAffineEstimation(srcPts(goodIdxs,:), dstPts(goodIdxs,:), invErrCovs(:, :, goodIdxs), srcPts(goodIdxs,:));
                    [ ~, mahalanobisErrs ] = evalWeightedAffineEstimation(Hhat, predictabilityCovs, srcPts(goodIdxs,:), dstPts(goodIdxs,:), errCovs(:,:,goodIdxs));
                    goodIdxs = mahalanobisErrs < (this.timesSigmaPredictionSlackFactor)/2;
                    if (all(goodIdxs))
                        isSucsses = true;
                    else
                        isSucsses = false;
                        chosenIdxs = [];
                    end
                    
                    
                else
                    isSucsses = false;
                    chosenIdxs = [];
                end
            end
            this.currentCorrespondences.src = [];
            this.currentCorrespondences.dst = [];
            this.currentCorrespondences.qualities = [];
            this.currentCorrespondences.covMats = [];
            this.currentCorrespondences.invCovMats = [];
            if (isSucsses)
                this.currentCorrespondences.src = srcPts(chosenIdxs,:);
                this.currentCorrespondences.dst = dstPts(chosenIdxs,:);
                this.currentCorrespondences.qualities = qualities(chosenIdxs);
                this.currentCorrespondences.covMats = errCovs(:,:,chosenIdxs);
                this.currentCorrespondences.invCovMats = invErrCovs(:,:,chosenIdxs);
                [this.srcPtsCentroid, v1, v2, ~, P] = getRegionPrincipleAxes(this.currentCorrespondences.src);
                if (P ~= 0)
                    this.wasStopped = true;
                    this.wasExpanded = false;
                    return;
                end
                this.srcPrincipalAxes(:,1) = v1;
                this.srcPrincipalAxes(:,2) = v2;
                [~, v1, v2, ~, ~] = getRegionPrincipleAxes(this.currentCorrespondences.dst);
                this.dstPrincipalAxes(:,1) = v1;
                this.dstPrincipalAxes(:,2) = v2;
                [this.wasExpanded, Hhat] = this.isStatisticallyExpandedEnough;
                this.Hprevious = this.Hhat;
                this.Hhat = Hhat;
                this.currentCorrespondences.dst = doAffineTransform(this.currentCorrespondences.src, Hhat);
                this.area = pi*norm(this.srcPrincipalAxes(:,1))*norm(this.srcPrincipalAxes(:,2));
            else
                this.wasRejected = true;
                this.wasStopped = true;
            end
        end
        function [wasSucsses, templates, dstPts, srcPts, actualExpansionFactor] = prepareInitialSamples(this, srcImg, srcCornerImg, expansionFactor)
            % prepareInitialSamples - get samples according to the desired expansion 
            % Inputs - see initialExpansion
            % Outputs:
            %   wasSucsses - flag indicating whether the sampling sucsseeded
            %   templates - the extracted templates(patches) around the samples
            %   dstPts - the predicted locations in the destination image
            %   srcPts - the samples in the source image
            %   actualExpansionFactor - the expansion factor in practive after possible reductions
            wasSucsses = false;
            desiredProjectionPoint = this.srcPtsCentroid + expansionFactor * this.srcPrincipalAxes(:,2)';
            X = zeros(2,6);
            X(1,1:3) = [desiredProjectionPoint 1];
            X(2,4:6) = [desiredProjectionPoint 1];
            xySlacks = sqrt(diag(X / this.invHerrCovariance *(X'))) * this.timesSigmaPredictionSlackFactor;
            expansionReduction = min(this.maximalSearchWindowSlack / max(xySlacks), 1);
            actualExpansionFactor = expansionFactor*expansionReduction;
            [sampleEllipse] = this.determineExpansionEllipse(actualExpansionFactor);
            prescribedNumberOfSamples = max(ceil(pi*(sqrt(prod(sum(sampleEllipse.^2, 1))))*this.downSamplingFactor^2), this.minimalInitialSamples);
            numberOfLayers = ceil(log2(prescribedNumberOfSamples/regionExpanderDeltaBased.minimalThetaSampleRes + 1));
            [unitCircularPoints]  = regionExpanderDeltaBased.sampleCircle(1, numberOfLayers, regionExpanderDeltaBased.minimalThetaSampleRes);
            srcPts = (sampleEllipse*unitCircularPoints')';%still around the origin
            srcPts = [(this.srcPtsCentroid(1) + srcPts(:,1)), (this.srcPtsCentroid(2) + srcPts(:,2))];%offset to around centroid
            
            %shifting is instead of throwing due to OOB
            shiftLeft  = max(srcPts(:,1) - (size(srcImg, 2) - this.templateExtent));shiftLeft = (shiftLeft+1) * (shiftLeft > 1);
            shiftRight = min(srcPts(:,1) - this.templateExtent);shiftRight = (shiftRight-1) * (shiftRight < 1);
            shiftUp = max(srcPts(:,2) - (size(srcImg, 1) - this.templateExtent)); shiftUp = (shiftUp+1) * (shiftUp > 1);
            shiftDown = min(srcPts(:,2) - this.templateExtent);shiftDown = (shiftDown-1) * (shiftDown < 1);
            srcPts = [srcPts(:,1) - shiftLeft - shiftRight, srcPts(:,2) - shiftDown - shiftUp];
            oobIndicator = srcPts(:,1) > (size(srcImg, 2) - this.templateExtent - 1) | srcPts(:,1) < this.templateExtent + 1 |...
                srcPts(:,2) > (size(srcImg, 1) - this.templateExtent - 1) | srcPts(:,2) < this.templateExtent + 1;
            srcPts = srcPts(~oobIndicator,:);
            if (size(srcPts, 1) < 4)    %minimum of 4 points is required to estimated and verify the affine transformation
                templates = [];
                dstPts = [];
                srcPts = [];
                return;
            end
            tollerance = floor(sqrt(det(sampleEllipse)/size(srcPts, 1)));
            pointsAdjustedFlag = false;
            if (~isempty(srcCornerImg) && tollerance > (this.templateExtent/2) )
                pointsAdjustedFlag = true;
                tollerance = min(tollerance, this.templateExtent-1);
                [adjustedSrcPts, qualities] = adjustToBestCloseCorner(srcCornerImg, srcPts, [tollerance tollerance]);
                oobIndicator = adjustedSrcPts(:,1) > (size(srcImg, 2) - this.templateExtent - 1) | adjustedSrcPts(:,1) < this.templateExtent + 1 |...
                    adjustedSrcPts(:,2) > (size(srcImg, 1) - this.templateExtent - 1) | adjustedSrcPts(:,2) < this.templateExtent + 1;
                srcPts(~oobIndicator,:) = adjustedSrcPts(~oobIndicator,:);
                [~, idxs] = sort(qualities, 'Descend');
                srcPts = srcPts(idxs,:);
            end
            if size(srcPts,1) > regionExpanderDeltaBased.maximalNumberOfInitialSamples
                if(pointsAdjustedFlag)
                   idxs = 1:regionExpanderDeltaBased.maximalNumberOfInitialSamples;
                else
                   idxs = randsample(size(srcPts, 1), regionExpanderDeltaBased.maximalNumberOfInitialSamples);
                end
                srcPts = srcPts(idxs,:);
            end
            %Throw redundant (too close) points on both source and destination
            if (exist('uniquetol','builtin'))
                dstPts = doAffineTransform(srcPts,this.Hhat);
                [~,selectedIdxsSrc] = uniquetol(srcPts,(this.templateExtent/2),'ByRows',true,'DataScale', 1,'OutputAllIndices',true);
                selectedIdxsSrc = cellfun(@(v) v(1), selectedIdxsSrc(:,1));
                [~,selectedIdxsDst] = uniquetol(dstPts,(this.templateExtent/2),'ByRows',true,'DataScale', 1,'OutputAllIndices',true);
                selectedIdxsDst = cellfun(@(v) v(1), selectedIdxsDst(:,1));
                selectedIdxs = intersect(selectedIdxsSrc,selectedIdxsDst);
                srcPts = srcPts(selectedIdxs,:);
                if (size(srcPts, 1) < 4)    %minimum of 4 points is required to estimated and verify the affine transformation
                    templates = [];
                    dstPts = [];
                    srcPts = [];
                    return;
                end
            end
            if isequal(this.Hhat, eye(3))
                [templates] = getEyeTemplates(srcImg, srcPts, this.templateExtent);
                dstPts = srcPts;
                badIdxs = false(length(srcPts), 1);
            else
                [templates, dstPts, badIdxs] = getTemplates(srcImg, srcPts, this.templateExtent, this.Hhat);
            end
            templates = templates(~badIdxs);
            dstPts = dstPts(~badIdxs,:);
            srcPts = srcPts(~badIdxs,:);
            if (length(srcPts) >= 4)
                wasSucsses = true;
            end
        end
        function [bool, Hhat] = isStatisticallyExpandedEnough(this)
            % isStatisticallyExpandedEnough - check whether the region has enough statistical power to be considered verified
            % Outputs
            %   bool  - logical answer
            %   Hhat - the estimated affine transformation
            
            % Source to target check
            pointsAroundUs = repmat(this.srcPtsCentroid, 4, 1) + [this.minimalMinorAxisLength 0;...
                -this.minimalMinorAxisLength 0;...
                0 this.minimalMinorAxisLength;...
                0 -this.minimalMinorAxisLength;];
            [Hhat, this.invHerrCovariance, predicitabilityCovs, ~] = weighteAffineEstimation(...
                this.currentCorrespondences.src, this.currentCorrespondences.dst,...
                this.currentCorrespondences.invCovMats, pointsAroundUs);
            maximalPredictionVariances = arrayfun(@(x) max(eig(predicitabilityCovs(:,:,x))) ,1:4);
            boolSrc = all(maximalPredictionVariances < (this.maximalRelativePredictionErr * this.minimalMinorAxisLength)^2);
            
            % Target to source check
            pointsAroundUs = repmat(mean( this.currentCorrespondences.dst,1), 4, 1) + [this.minimalMinorAxisLength 0;...
                -this.minimalMinorAxisLength 0;...
                0 this.minimalMinorAxisLength;...
                0 -this.minimalMinorAxisLength;];
            [~, ~, predicitabilityCovs, ~] = weighteAffineEstimation(...
                this.currentCorrespondences.dst, this.currentCorrespondences.src,...
                this.currentCorrespondences.invCovMats, pointsAroundUs);
            maximalPredictionVariances = arrayfun(@(x) max(eig(predicitabilityCovs(:,:,x))) ,1:4);
            boolDst = all(maximalPredictionVariances < (this.maximalRelativePredictionErr * this.minimalMinorAxisLength)^2);
            bool = boolSrc && boolDst;
        end
                        
        %% further expansion
        function [] = furtherExpand(this, srcImg, dstImg, srcCornerImg, expansionFactor, isDebug)
            % furtherExpand - expansion of an already verified region
            % Inputs:
             %  srcImg - source image
             %  dstImg - destination image
             %  srcCornerImg - corrnerness map of the source image
             %  expansionFactor - how much to expand in every axis
             %  isDebug - a boolean flag  
            [isSuccess, dstRois, dstPts, srcPts] = this.prepareFurtherSamples(srcImg, srcCornerImg, expansionFactor);
            if (~isSuccess)
                this.wasStopped = true;
                return;
            end
            [dstPts, matchQualities, ~, errorCorrMats, isGarbageIndicators, predictabilityCovs] = this.scanForMatches(dstImg,dstRois, dstPts, srcPts, isDebug);
            srcPts(isGarbageIndicators,:) = [];
            dstPts(isGarbageIndicators,:) = [];
            matchQualities(isGarbageIndicators,:) = [];
            errorCorrMats(:,:,isGarbageIndicators) = [];
            predictabilityCovs(:,:,isGarbageIndicators) = [];
            goodQualitiesInd = matchQualities >  max([matchQualities;this.currentCorrespondences.qualities])*this.correlationDeviationTsh;
            if (any(goodQualitiesInd))
                previousArea = this.area;
                srcPts = srcPts(goodQualitiesInd,:);
                dstPts = dstPts(goodQualitiesInd,:);
                predictabilityCovs = predictabilityCovs(:,:,goodQualitiesInd);
                errorCorrMats = errorCorrMats(:,:,goodQualitiesInd);
                matchQualities = matchQualities(goodQualitiesInd);
                this.doGeometricAnalysisFurther(srcPts, dstPts, matchQualities, errorCorrMats, predictabilityCovs);
                if (this.area < (previousArea * this.minimalAreaGrowthFactor* expansionFactor))
                    this.wasStopped = true;
                end
            else
                this.wasStopped = true;
            end
        end
        
        function [] = doGeometricAnalysisFurther(this, srcPts, dstPts, newMatchQualities, newErrorCorrMats, predictabilityCovs)
            % doGeometricAnalysisFurther - check whether the given matches agree on an affine model produced by previous iterations
            % Inputs:
             %  srcPts - the new points in the source image
             %  dstPts - the new corresponding points in the destination image
             %  newMatchQualities - match qualites (correlations) of the given new matches
             %  predictabilityCovs - error covariance matrices when predicting the new matches
            [ ~, mahalanobisErrs ] = evalWeightedAffineEstimation(this.Hhat, predictabilityCovs, srcPts, dstPts, newErrorCorrMats);
            consensusTestInd = mahalanobisErrs < (this.timesSigmaPredictionSlackFactor/2);
            if ~any(consensusTestInd)
                this.wasStopped = true;
                return;
            end
            srcPts = srcPts(consensusTestInd,:);
            dstPts = dstPts(consensusTestInd,:);
            
            newMatchQualities = newMatchQualities(consensusTestInd);
            newErrorCorrMats = newErrorCorrMats(:,:,consensusTestInd);
            
            allSrcPts = [this.currentCorrespondences.src ; srcPts];
            allDstPts = [this.currentCorrespondences.dst ; dstPts];
            newErrorDets = zeros(size(srcPts, 1), 1);
            newInvErrCovs = zeros(size(newErrorCorrMats));
            qualityFactors = (newMatchQualities / max(newMatchQualities)).^-2;
            for k = 1:size(srcPts, 1)
                newErrorCorrMats(:,:,k) = qualityFactors(k) * newErrorCorrMats(:,:,k);
                [newInvErrCovs(:,:,k), newErrorDets(k)] = inv2By2Mat(newErrorCorrMats(:,:,k));
            end
            this.currentCorrespondences.covMats = cat(3, this.currentCorrespondences.covMats, newErrorCorrMats);
            this.currentCorrespondences.invCovMats = cat(3, this.currentCorrespondences.invCovMats, newInvErrCovs);
            this.Hprevious = this.Hhat;
            [this.Hhat, this.invHerrCovariance] = weighteAffineEstimation(allSrcPts, allDstPts, this.currentCorrespondences.invCovMats);
            % store
            this.currentCorrespondences.src = allSrcPts;
            this.currentCorrespondences.qualities = [this.currentCorrespondences.qualities;newMatchQualities];
            [this.srcPtsCentroid, v1, v2, ~, ~] = getRegionPrincipleAxes(this.currentCorrespondences.src);
            this.srcPrincipalAxes(:,1) = v1;
            this.srcPrincipalAxes(:,2) = v2;
            [~, v1, v2, ~, ~] = getRegionPrincipleAxes(this.currentCorrespondences.dst);
            this.dstPrincipalAxes(:,1) = v1;
            this.dstPrincipalAxes(:,2) = v2;
            this.currentCorrespondences.dst = doAffineTransform(this.currentCorrespondences.src, this.Hhat);
            this.area = pi*norm(this.srcPrincipalAxes(:,1))*norm(this.srcPrincipalAxes(:,2));
        end
        
        function [wasSucsses, templates, newDstPts, newSrcPts] = prepareFurtherSamples(this,srcImg,srcCornerImg, expansionFactor)
            % prepareFurtherSamples - get samples according to the desired expansion around matches for previous iterations 
            % Inputs - see furtherExpansion
            % Outputs:
            %   wasSucsses - flag indicating whether the sampling sucsseeded
            %   templates - the extracted templates(patches) around the samples
            %   newDstPts - the predicted locations in the destination image
            %   newSrcPts - the samples in the source image
            imageSize = size(srcImg);
            shortAxisLength = norm(this.srcPrincipalAxes(:,2));
            longAxisLength = norm(this.srcPrincipalAxes(:,1));
            newExtent = max([longAxisLength, shortAxisLength*expansionFactor]);
            [newSrcPts] = sampleAround(this.srcPtsCentroid, newExtent, this.downSamplingFactor, regionExpanderDeltaBased.minimalThetaSampleRes);
            X = zeros(2*size(newSrcPts, 1), 6);
            X(1:2:end,1:3) = [newSrcPts ones(size(newSrcPts, 1), 1)];
            X(2:2:end,4:6) = [newSrcPts ones(size(newSrcPts, 1), 1)];
            xySlacks = reshape(sqrt(diag(X / this.invHerrCovariance *(X'))) * this.timesSigmaPredictionSlackFactor, size(newSrcPts, 1), 2);
            expansionReductions = min(this.maximalSearchWindowSlack ./ max(xySlacks, [], 2), 1);
            C =  repmat(this.srcPtsCentroid, size(newSrcPts, 1), 1);
            newSrcPts = C + (newSrcPts - C).*repmat(expansionReductions, 1, 2);
            newSrcPts(newSrcPts(:,1) <=  (this.templateExtent+1)...
                | newSrcPts(:,2) <= (this.templateExtent+1)...
                | newSrcPts(:,1) > (imageSize(2)-this.templateExtent-1)...
                | newSrcPts(:,2) > (imageSize(1)-this.templateExtent-1), :) = [];
            if (isempty(newSrcPts))
                templates = [];
                newDstPts = [];
                newSrcPts = [];
                wasSucsses = false;
                return;
            end
            if (~isempty(srcCornerImg))
                [adjustedSrcPts, ~] = adjustToBestCloseCorner(srcCornerImg, newSrcPts, [this.templateExtent this.templateExtent]/2);
                oobIndicator = adjustedSrcPts(:,1) > (size(srcImg,2) - this.templateExtent - 1) | adjustedSrcPts(:,1) < this.templateExtent + 1 |...
                    adjustedSrcPts(:,2) > (size(srcImg, 1) - this.templateExtent - 1) | adjustedSrcPts(:,2) < this.templateExtent + 1;
                newSrcPts(~oobIndicator,:) = adjustedSrcPts(~oobIndicator,:);
            end
            [templates, newDstPts, badIdxs] = getTemplates(srcImg, newSrcPts, this.templateExtent, this.Hhat);
            templates = templates(~badIdxs);
            newDstPts = newDstPts(~badIdxs,:);
            newSrcPts = newSrcPts(~badIdxs,:);
            if (isempty(newSrcPts))
                wasSucsses = false;
            else
                wasSucsses = true;
            end
        end
        
        
        function [sampleEllipse] = determineExpansionEllipse(this, expansionFactor)
            % determineExpansionEllipse - determine the ellipse sample pattern of the expansion
            % Outputs:
            %   sampleEllipse - the matrix describing the ellispe (taking the canonic circle to the desired ellipse)
            plannedShortAxisLength = expansionFactor*norm(this.srcPrincipalAxes(:,2));
            plannedLongAxisLength = max(plannedShortAxisLength, norm(this.srcPrincipalAxes(:,1)));
            v1 = this.srcPrincipalAxes(:,1) / norm(this.srcPrincipalAxes(:,1));
            v2 = this.srcPrincipalAxes(:,2) / norm(this.srcPrincipalAxes(:,2));
            sampleEllipse = [plannedLongAxisLength*v1 plannedShortAxisLength*v2];
        end
        
  
        
end
    

%%
methods(Static)
    
    function [subImage, offset, wasFail] = cropSubImage(img, centerPnt, searchExtentXY, minimalExtent)
        %cropSubImage - crop a sub image with a given extent (searchExtentXY) around <centerPnt>
        % Inputs:
            % img- the given (whole) image
            % centerPnt - the point to crop around
            % searchExtentXY - the extent of the cropped rectangle (a 2 vector)
            % minimalExtent - the minimal extent acceptable on both axes
        % Outputs:
            % subImage - the cropped image
            % offset - coordinate system offest of the subImage relative to the original image
            % wasFail - could we crop at least <minimalExtent> around <centerPnt>
        wasFail = false;
        yIdxs = round((centerPnt(2) - searchExtentXY(2)):(centerPnt(2) + searchExtentXY(2)));
        xIdxs = round((centerPnt(1) - searchExtentXY(1)):(centerPnt(1) + searchExtentXY(1)));
        xLeftOOB = sum(xIdxs < 1);xRightOOB = sum((xIdxs > size(img, 2)));xCut = max(xLeftOOB,xRightOOB);
        yUpOOB = sum(yIdxs < 1);yDownOOB = sum((yIdxs > size(img, 1)));yCut = max(yUpOOB,yDownOOB);
        xIdxs = xIdxs((1+xCut):(end - xCut));
        yIdxs = yIdxs((1+yCut):(end - yCut));
        if (length(xIdxs) < (minimalExtent*2 + 1) || length(yIdxs) < (minimalExtent*2 + 1))
            wasFail = true;
            subImage =[];offset=[];
            return;
        end
        subImage = img(yIdxs, xIdxs);
        offset = [xIdxs(1), yIdxs(1)];
    end
    
    
    
    %visualzation related
    function [] = textifyMyPoints(XYpts, color)
        a = (1:length(XYpts))'; b = num2str(a); c = cellstr(b);
        dx = 1; dy = 1; % displacement so the text does not overlay the data points
        text(double(XYpts(:,1)+dx), double(XYpts(:,2) + dy), c, 'Color', color, 'FontWeight', 'bold', 'FontSize', 16, 'Tag', 'textTag');
    end
    
    %%
    function [points]  = sampleCircle(radius, levels, initialThetaRes) 
        % sampleCircle - get 2D samples on a circle (or circles) around the origin
        % Inputs:
            % radius- the radius of the outermost circle (level)
            % levels - how many concentric circles
            % initialThetaRes - amount of samples on the smalles circle
        % Outputs:
            % points - the samples
        points = zeros(initialThetaRes*(2^levels -1), 2);
        currentRadius = radius / (2^(levels-1));
        currentAmountOfThetas = initialThetaRes;
        currentIndex = 1;
        for levelIdx = 1:1:levels
            thetaShift = mod(levelIdx*pi/6, pi/2);%(2*pi)/(2*currentAmountOfThetas);
            points(currentIndex:(currentIndex+currentAmountOfThetas-1),:) = [currentRadius*ones(currentAmountOfThetas,1), linspace(0, 2*pi*(1 - 1/currentAmountOfThetas), currentAmountOfThetas)' + thetaShift];
            currentIndex = currentIndex + currentAmountOfThetas;
            currentAmountOfThetas = currentAmountOfThetas * 2;
            %                 currentRadius = currentRadius * (sqrt(2));
            currentRadius = currentRadius * 2;
        end
        [points(:,1), points(:,2)] = pol2cart(points(:,2), points(:,1));
        
    end
    
    
end




end