function [cords, maxCorr, errorCorrMat, isGarbage, shiftTip, initialCorrelation] = patchedFilterDelta(dstImg, template, isDebug, corrTsh)
% Normalized Cross Correlation based matched filter
% 
% looks for the maximal correlation of template in dstImg.
% returns several outputs corresponding the correlation between dstImg and
% template.
% 
% input:            dstImg, template - two gray scale images
%                   corrTsh          - a minimum thresh hold for maximum
%                                      correlation value
% 
% output:           cords            - maximum value coordinates
%                   maxCorr          - maximum value
%                   errorCorrMat     - error covariance matrix
%                   isGarbage        - boolean for flagging a non acceptable
%                                      result as a correlation result
%                   shiftTip         - A suggestion on how to move the center of the scanninf window in case a re-scan is desired.                                                         

QE = 0.75;
isGarbage = false;
shiftTip = [0 0];
desiredPrecision = 5;               % maximum distance between maximal value and the farthest value within corrTsh
cords = floor(flip(size(dstImg)/2));
maxCorr = 0;
errorCorrMat = inf(2);
initialCorrelation = 0;

if (sum(abs(template(:) - mean(template(:)))) <= 3*numel(template)) % rejecting homogeneous patch for lack of information
    %        disp('Dumping homogenic patch');
    if(isDebug)
        figure(93);
        imshow(template,[]);
        title('template')
    end
    isGarbage = true;
    return;
end
templateSize = size(template); 
isOdd = bitget(templateSize, 1);
templateExtent = (templateSize - isOdd)/2;
C = normxcorr2_mex(double(template), double(dstImg), 'valid');
responseSize = size(C);
initialCorrelation = C(round(responseSize(1)/2),round(responseSize(2)/2));
[maxCorr, linIdx] = max(C(:));

if (maxCorr < corrTsh)  % rejecting correlation results where there is no value over the minimal value that was set by the user
    isGarbage = true;
   return;
end

[maxCord(2), maxCord(1)] =ind2sub(responseSize, linIdx);
[alternativesY, alternativesX] = find(C >= QE*maxCorr);
[radius] = getMaxDistanceFromPoint([alternativesX(:) alternativesY(:)], maxCord);
[cords] = maxCord + flip(templateExtent - isOdd);
if (isDebug)
    %%
    figure(94);
    subplot(1,2,1)
    imshow(C, []);hold on
    title('filter response');
    scatter(alternativesX, alternativesY, 'b', 'fill')
    scatter(maxCord(1), maxCord(2), 'g', 'fill');
    subplot(1,2,2)
    imshow(dstImg, []);hold on
    scatter(maxCord(1) - isOdd(1) + templateExtent(2), maxCord(2) - isOdd(2) + templateExtent(1), 'g', 'fill');
    rectangle('Position', [maxCord(1)-isOdd(1), maxCord(2) - isOdd(2), templateSize(2), templateSize(1)], 'LineWidth', 4, 'EdgeColor', 'g')
    title('target image');
    figure(95); imshow(template, []); title('template');
    %%
end
if (radius > desiredPrecision)      % rejecting correlation results due to too far away high correlation values
    isGarbage = true;
    return;
end
    
[XX,YY] = meshgrid(1:responseSize(2), 1:responseSize(1));
XX(C < maxCorr*QE) = [];XX = XX(:);
YY(C < maxCorr*QE) = [];YY = YY(:);
C(C < maxCorr*QE) = [];C = C(:);

if (XX(1) == 1 || XX(end) == responseSize(2) || any(YY == 1) || any(YY == responseSize(1))) % 
   numberOfFrameTouches = sum((XX == 1) + (XX == responseSize(2)) + (YY == 1) + (YY == responseSize(1)));
   covarianceGrowthFactor = 2*(numberOfFrameTouches/sum(responseSize)) + 1;
   shiftTip = maxCord - flip(responseSize)/2 - isOdd;
else
   covarianceGrowthFactor = 1;
end

C = C/sum(C);       % we treat the correlation result as a distribution so we could analyze it statistically
muX = C'*XX;
muY = C'*YY;
XX = XX - muX;
YY = YY - muY;
varX = max(C'*(XX.^2), 0.25);
varY = max(C'*(YY.^2), 0.25);
covXY = (C'*(XX.*YY));
covXY = covXY * (covXY^2 < varX * varY);%force positive definite
errorCorrMat = [varX covXY; covXY varY] * covarianceGrowthFactor^2;
if (isDebug)
    figure(94);
    subplot(1,2,1)
    hold on; h = error_ellipse(errorCorrMat,[muX muY])  ;set(h, 'color', 'r', 'LineWidth', 3);
end
if (isDebug)
    figure(94);subplot(1,2,2);scatter(cords(1), cords(2), 'r', 'fill')
    maxCorr
    shiftTip
%     quality
%     pause
    close all
end

end


% aid function for maximum distance between one point and a group of points
function [distance] = getMaxDistanceFromPoint(group, point)
    dists = single(group) - repmat(point, size(group, 1), 1);
    dists = sum(dists.^2, 2);
    [val, ~] = max(dists);
    distance = sqrt(val);
end




